from typing import Any

import numpy as np
import numpy.typing as npt

def r_squared(x: npt.NDArray[np.float_], y: npt.NDArray[np.float_]) -> float: ...
def ols_swiping(x: npt.NDArray[np.float_], y: npt.NDArray[np.float_], kwargs: Any) -> int: ...



from typing import Any

import numpy as np
import numpy.typing as npt

def get_knee(c: float) -> float: ...
def c_method(x: npt.NDArray[np.float_], y: npt.NDArray[np.float_], method: str, kwargs: Any) -> int: ...



# SPDX-FileCopyrightText: 2021-present InCogNiTo124 <msmetko@msmetko.xyz>
#
# SPDX-License-Identifier: Apache-2.0

from collections import Counter
import enum
from functools import partial
from pathlib import Path
from typing import List, Optional

import typer

from knarrow import find_knee

from ..__about__ import __version__


class Method(str, enum.Enum):
    ALL = "all"
    ANGLE = "angle"
    C_METHOD = "c_method"
    DISTANCE = "distance"
    DISTANCE_ADJACENT = "distance_adjacent"
    KNEEDLE = "kneedle"
    MENGER_ANCHORED = "menger_anchored"
    MENGER_SUCCESSIVE = "menger_successive"
    OLS_SWIPING = "ols_swiping"


class Output(str, enum.Enum):
    INDEX = "index"
    VALUE = "value"


app = typer.Typer()


def stdin_callback(value: Optional[Path]) -> List[Path]:
    return value if value else [Path("/dev/stdin")]


def version_callback(value: bool) -> None:
    if value:
        print(f"knarrow v{__version__}\nType `knarrow --help` for further info.")
        raise typer.Exit(0)


@app.command()
def main(
    method: Method = typer.Option("all", help="The method to use to calculate the knee's position"),
    files: List[Path] = typer.Argument(
        None,
        allow_dash=True,
        exists=True,
        dir_okay=False,
        readable=True,
        callback=stdin_callback,
        help="List of input files (default: stdin)",
        show_default=False,
    ),
    sort: bool = typer.Option(False, help="Whether or not to sort the input"),
    delimiter: str = typer.Option(
        ",", help="If the input is 2-dimensional, split the dimensions by this option's value"
    ),
    output: Output = typer.Option(
        "value", help="Type of output. Value means value itself, index means it's ordinal number"
    ),
    smoothing: float = typer.Option(0.0, min=0.0, help="Cubic spline smoothing parameter"),
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
    ),  # version boilerplate
):
    for path in files:
        with path.open("r") as file:
            rows = list(map(str.strip, file))
            split = partial(str.split, sep=delimiter)
            values = map(split, rows)
            numbers = list(tuple(float(value) for value in row) for row in values)
            indices = list(range(len(numbers)))
            if sort:
                indices.sort(key=lambda i: numbers[i])
                key_function = (lambda x: x) if len(numbers[0]) == 1 else (lambda x: x[0])
                numbers.sort(key=key_function)

            if method == Method.ALL:
                counter = Counter(
                    [
                        find_knee(numbers, method=m.value, sort=sort, smoothing=smoothing)
                        for m in Method
                        if m != Method.ALL
                    ]
                )
                most_common = counter.most_common(1).pop(0)
                knee = most_common[0]
            else:
                knee = find_knee(numbers, method=method.value, sort=sort, smoothing=smoothing)

            result = indices[knee] if output == Output.INDEX else rows[indices[knee]]
            print(path.name, result)
    return



from typing import Any

import numpy as np
import numpy.typing as npt

def menger_successive(x: npt.NDArray[np.float_], y: npt.NDArray[np.float_], kwargs: Any) -> int: ...
def menger_anchored(x: npt.NDArray[np.float_], y: npt.NDArray[np.float_], kwargs: Any) -> int: ...
def double_triangle_area(vertices: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]: ...
def get_squared_vector_lengths(vertices: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]: ...
def get_curvature(vertices: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]: ...



import enum
from typing import Any, Callable, Tuple

import numpy as np
import numpy.typing as npt

class KneeType(enum.Enum):
    DECREASING_CONVEX = 0
    INCREASING_CONCAVE = 1
    DECREASING_CONCAVE = 2
    INCREASING_CONVEX = 3

def np_windowed(length: int, window_size: int, stride: int = 1, dilation: int = 1) -> npt.NDArray[np.int_]: ...
def np_anchored(length: int) -> npt.NDArray[np.int_]: ...
def prepare(f: Callable[[Any], int]) -> int: ...
def normalize(x: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]: ...
def get_delta_matrix(x: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]: ...
def get_weight_matrix(x: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]: ...
def cubic_spline_smoothing(
    x: npt.NDArray[np.float_], y: npt.NDArray[np.float_], smoothing_factor: float
) -> Tuple[npt.NDArray[np.float_], npt.NDArray[np.float_]]: ...
def projection_distance(vertices: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]: ...
def detect_knee_type(y1: float, y2: float, y3: float, y4: float) -> KneeType: ...



from typing import Any

import numpy as np
import numpy.typing as npt

def distance(x: npt.NDArray[np.float_], y: npt.NDArray[np.float_], kwargs: Any) -> int: ...
def distance_adjacent(x: npt.NDArray[np.float_], y: npt.NDArray[np.float_], kwargs: Any) -> int: ...
